"""Definition of the audio service backends base classes.

These classes can be used to create an Audioservice plugin extending
OpenVoiceOS's media playback options.
"""
from abc import ABCMeta, abstractmethod
from typing import List

from ovos_bus_client import Message
from ovos_bus_client.message import dig_for_message
from ovos_utils import classproperty
from ovos_utils.fakebus import FakeBus
from ovos_utils.log import LOG
from ovos_utils.process_utils import RuntimeRequirements

try:
    from ovos_utils.ocp import PlaybackType, TrackState, PlayerState, MediaState
except ImportError:
    LOG.warning("Please update to ovos-utils~=0.1.")
    from enum import IntEnum


    class MediaState(IntEnum):
        # https://doc.qt.io/qt-5/qmediaplayer.html#MediaStatus-enum
        # The status of the media cannot be determined.
        UNKNOWN = 0
        # There is no current media. PlayerState == STOPPED
        NO_MEDIA = 1
        # The current media is being loaded. The player may be in any state.
        LOADING_MEDIA = 2
        # The current media has been loaded. PlayerState== STOPPED
        LOADED_MEDIA = 3
        # Playback of the current media has stalled due to
        # insufficient buffering or some other temporary interruption.
        # PlayerState != STOPPED
        STALLED_MEDIA = 4
        # The player is buffering data but has enough data buffered
        # for playback to continue for the immediate future.
        # PlayerState != STOPPED
        BUFFERING_MEDIA = 5
        # The player has fully buffered the current media. PlayerState != STOPPED
        BUFFERED_MEDIA = 6
        # Playback has reached the end of the current media. PlayerState == STOPPED
        END_OF_MEDIA = 7
        # The current media cannot be played. PlayerState == STOPPED
        INVALID_MEDIA = 8


    class PlayerState(IntEnum):
        # https://doc.qt.io/qt-5/qmediaplayer.html#State-enum
        STOPPED = 0
        PLAYING = 1
        PAUSED = 2


    class PlaybackType(IntEnum):
        SKILL = 0  # skills handle playback whatever way they see fit,
        # eg spotify / mycroft common play
        VIDEO = 1  # Video results
        AUDIO = 2  # Results should be played audio only
        AUDIO_SERVICE = 3  ## DEPRECATED - used in ovos 0.0.7
        MPRIS = 4  # External MPRIS compliant player
        WEBVIEW = 5  # webview, render a url instead of media player
        UNDEFINED = 100  # data not available, hopefully status will be updated soon..


    class TrackState(IntEnum):
        DISAMBIGUATION = 1  # media result, not queued for playback
        PLAYING_SKILL = 20  # Skill is handling playback internally
        PLAYING_AUDIOSERVICE = 21  ## DEPRECATED - used in ovos 0.0.7
        PLAYING_VIDEO = 22  # Skill forwarded playback to video service
        PLAYING_AUDIO = 23  # Skill forwarded playback to audio service
        PLAYING_MPRIS = 24  # External media player is handling playback
        PLAYING_WEBVIEW = 25  # Media playback handled in browser (eg. javascript)

        QUEUED_SKILL = 30  # Waiting playback to be handled inside skill
        QUEUED_AUDIOSERVICE = 31  ## DEPRECATED - used in ovos 0.0.7
        QUEUED_VIDEO = 32  # Waiting playback in video service
        QUEUED_AUDIO = 33  # Waiting playback in audio service
        QUEUED_WEBVIEW = 34  # Waiting playback in browser service


class AudioBackend(metaclass=ABCMeta):
    """Base class for all audio backend implementations.

    Arguments:
        config (dict): configuration dict for the instance
        bus (MessageBusClient): OpenVoiceOS messagebus emitter
    """

    def __init__(self, config=None, bus=None, name=None):
        self.name = name or self.__class__.__name__
        self._now_playing = None  # single uri
        self._tracks = []  # list of dicts for OCP entries
        self._idx = 0
        self._track_start_callback = None
        self.supports_mime_hints = False
        self.config = config or {}
        self.bus = bus or FakeBus()

    @classproperty
    def runtime_requirements(self):
        """ skill developers should override this if they do not require connectivity
         some examples:
         IOT plugin that controls devices via LAN could return:
            scans_on_init = True
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=scans_on_init,
                                 requires_internet=False,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=False)
         online search plugin with a local cache:
            has_cache = False
            RuntimeRequirements(internet_before_load=not has_cache,
                                 network_before_load=not has_cache,
                                 requires_internet=True,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
         a fully offline plugin:
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=False,
                                 requires_internet=False,
                                 requires_network=False,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
        """
        return RuntimeRequirements(internet_before_load=False,
                                   network_before_load=False,
                                   requires_internet=False,
                                   requires_network=False,
                                   no_internet_fallback=True,
                                   no_network_fallback=True)

    @property
    def playback_time(self) -> int:
        """ in milliseconds """
        return 0

    @abstractmethod
    def supported_uris(self) -> List[str]:
        """List of supported uri types.

        Returns:
            list: Supported uri's
        """

    @abstractmethod
    def play(self, repeat=False):
        """Start playback.

        Starts playing the first track in the playlist and will contiune
        until all tracks have been played.

        Arguments:
            repeat (bool): Repeat playlist, defaults to False
        """

    @abstractmethod
    def lower_volume(self):
        """Lower volume.

        This method is used to implement audio ducking. It will be called when
        OpenVoiceOS is listening or speaking to make sure the media playing isn't
        interfering.
        """

    @abstractmethod
    def restore_volume(self):
        """Restore normal volume.

        Called when to restore the playback volume to previous level after
        OpenVoiceOS has lowered it using lower_volume().
        """

    @abstractmethod
    def get_track_length(self) -> int:
        """
        getting the duration of the audio in miliseconds
        """

    @abstractmethod
    def get_track_position(self) -> int:
        """
        get current position in miliseconds
        """

    @abstractmethod
    def set_track_position(self, milliseconds):
        """Go to X position.
        Arguments:
           milliseconds (int): position to go to in milliseconds
        """

    @abstractmethod
    def pause(self):
        """Pause playback.

        Stops playback but may be resumed at the exact position the pause
        occured.
        """

    @abstractmethod
    def resume(self):
        """Resume paused playback.

        Resumes playback after being paused.
        """

    @abstractmethod
    def stop(self):
        """Stop playback.

        Stops the current playback.

        Returns:
            bool: True if playback was stopped, otherwise False
        """

    #####################
    # internals and default implementations
    def track_info(self) -> dict:
        """Request information of current playing track.
        Returns:
            Dict with track info.
        """
        return self._uri2meta(self._now_playing)

    def clear_list(self):
        """Clear playlist."""
        self._tracks = []
        self._idx = 0

    def add_list(self, tracks):
        """Add tracks to backend's playlist.

        Arguments:
            tracks (list): list of tracks.
        """
        tracks = tracks or []
        if isinstance(tracks, str):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        if tracks and not self._tracks:
            self.load_track(tracks[0])
            self._idx = 0
        self._tracks += tracks

    def next(self):
        """Skip to next track in playlist."""
        self._idx += 1
        if self._idx < len(self._tracks):
            self.load_track(self._tracks[self._idx])
            self.play()
        else:
            LOG.error("no more tracks!")

    def previous(self):
        """Skip to previous track in playlist."""
        self._idx = max(self._idx - 1, 0)
        if self._idx < len(self._tracks):
            self.load_track(self._tracks[self._idx])
            self.play()
        else:
            LOG.error("already in first track!")

    def seek_forward(self, seconds=1):
        """Skip X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() + miliseconds
        self.set_track_position(new_pos)

    def seek_backward(self, seconds=1):
        """Rewind X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative jump forward.
        """
        miliseconds = seconds * 1000
        new_pos = self.get_track_position() - miliseconds
        self.set_track_position(new_pos)

    def set_track_start_callback(self, callback_func):
        """Register callback on track start.

        This method should be called as each track in a playlist is started.
        """
        self._track_start_callback = callback_func

    def shutdown(self):
        """Perform clean shutdown.

        Implements any audio backend specific shutdown procedures.
        """
        self.stop()

    def _format_msg(self, msg_type, msg_data=None):
        # this method ensures all skills are .forward from the utterance
        # that triggered the skill, this ensures proper routing and metadata
        msg_data = msg_data or {}
        msg = dig_for_message()
        if msg:
            msg = msg.forward(msg_type, msg_data)
        else:
            msg = Message(msg_type, msg_data)
        # at this stage source == skills, lets indicate audio service took over
        sauce = msg.context.get("source")
        if sauce == "skills":
            msg.context["source"] = "audio_service"
        return msg

    ############################
    # OCP extensions - new methods to improve compat with OCP
    @staticmethod
    def _uri2meta(uri):
        if isinstance(uri, list):
            uri = uri[0]
        try:
            from ovos_ocp_files_plugin.plugin import OCPFilesMetadataExtractor
            return OCPFilesMetadataExtractor.extract_metadata(uri)
        except:
            meta = {"uri": uri,
                    "skill_id": "mycroft.audio_interface",
                    "playback": PlaybackType.AUDIO,  # TODO mime type check
                    "status": TrackState.QUEUED_AUDIO,
                    }
        return meta

    def load_track(self, uri):
        """ This method is only used by ovos-core
        In ovos audio backends are single-track, playlists are handled by OCP
        """
        self._now_playing = uri
        LOG.debug(f"queuing for {self.__class__.__name__} playback: {uri}")
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADING_MEDIA}))
        self.bus.emit(Message("ovos.common_play.track.state", {
            "state": TrackState.QUEUED_AUDIOSERVICE
        }))

    def ocp_sync_playback(self, playback_time):
        self.bus.emit(Message("ovos.common_play.playback_time",
                              {"position": playback_time,
                               "length": self.get_track_length()}))

    def ocp_start(self):
        """Emit OCP status events for play"""
        self.bus.emit(Message("ovos.common_play.player.state",
                              {"state": PlayerState.PLAYING}))
        self.bus.emit(Message("ovos.common_play.media.state",
                              {"state": MediaState.LOADED_MEDIA}))
        self.bus.emit(Message("ovos.common_play.track.state",
                              {"state": TrackState.PLAYING_AUDIOSERVICE}))

    def ocp_error(self):
        """Emit OCP status events for playback error"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.INVALID_MEDIA}))
            self._now_playing = None

    def ocp_stop(self):
        """Emit OCP status events for stop"""
        if self._now_playing:
            self._now_playing = None
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))
            self.bus.emit(Message("ovos.common_play.media.state",
                                  {"state": MediaState.END_OF_MEDIA}))

    def ocp_pause(self):
        """Emit OCP status events for pause"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PAUSED}))

    def ocp_resume(self):
        """Emit OCP status events for resume"""
        if self._now_playing:
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PLAYING}))
            self.bus.emit(Message("ovos.common_play.track.state",
                                  {"state": TrackState.PLAYING_AUDIOSERVICE}))


class RemoteAudioBackend(AudioBackend):
    """Base class for remote audio backends.

    RemoteAudioBackends will always be checked after the normal
    AudioBackends to make playback start locally by default.

    An example of a RemoteAudioBackend would be things like Chromecasts,
    mopidy servers, etc.
    """



from ovos_plugin_manager.utils import normalize_lang, \
    PluginTypes, PluginConfigTypes
from ovos_plugin_manager.templates.solvers import QuestionSolver, TldrSolver, \
    EntailmentSolver, MultipleChoiceSolver, EvidenceSolver
from ovos_utils.log import LOG



def find_question_solver_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.QUESTION_SOLVER)


def load_question_solver_plugin(module_name: str) -> type(QuestionSolver):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.QUESTION_SOLVER)


def get_question_solver_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.QUESTION_SOLVER)


def get_question_solver_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: lists of dict configurations by language (if provided)
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    # solver plugins return {lang: [list of config dicts]}
    return load_plugin_configs(module_name, PluginConfigTypes.QUESTION_SOLVER,
                               True)


def get_question_solver_lang_configs(lang: str,
                                     include_dialects: bool = False) -> dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.QUESTION_SOLVER, lang,
                                       include_dialects)


def get_question_solver_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.QUESTION_SOLVER)


def find_tldr_solver_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.TLDR_SOLVER)


def load_tldr_solver_plugin(module_name: str) -> type(TldrSolver):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.TLDR_SOLVER)


def get_tldr_solver_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.TLDR_SOLVER)


def get_tldr_solver_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: lists of dict configurations by language (if provided)
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    # solver plugins return {lang: [list of config dicts]}
    return load_plugin_configs(module_name, PluginConfigTypes.TLDR_SOLVER, True)


def get_tldr_solver_lang_configs(lang: str,
                                 include_dialects: bool = False) -> dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.TLDR_SOLVER, lang,
                                       include_dialects)


def get_tldr_solver_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.TLDR_SOLVER)


def find_entailment_solver_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.ENTAILMENT_SOLVER)


def load_entailment_solver_plugin(module_name: str) -> type(EntailmentSolver):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.ENTAILMENT_SOLVER)


def get_entailment_solver_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.ENTAILMENT_SOLVER)


def get_entailment_solver_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: dict configurations by language (if provided)
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    # solver plugins return {lang: [list of config dicts]}
    return load_plugin_configs(module_name, PluginConfigTypes.ENTAILMENT_SOLVER,
                               True)


def get_entailment_solver_lang_configs(lang: str,
                                     include_dialects: bool = False) -> dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.ENTAILMENT_SOLVER, lang,
                                       include_dialects)


def get_entailment_solver_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.ENTAILMENT_SOLVER)


def find_multiple_choice_solver_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.MULTIPLE_CHOICE_SOLVER)


def load_multiple_choice_solver_plugin(module_name: str) -> \
        type(MultipleChoiceSolver):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.MULTIPLE_CHOICE_SOLVER)


def get_multiple_choice_solver_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.MULTIPLE_CHOICE_SOLVER)


def get_multiple_choice_solver_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: dict configurations by language (if provided)
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    # solver plugins return {lang: [list of config dicts]}
    return load_plugin_configs(module_name,
                               PluginConfigTypes.MULTIPLE_CHOICE_SOLVER, True)


def get_multiple_choice_solver_lang_configs(lang: str,
                                            include_dialects: bool = False) -> \
        dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.MULTIPLE_CHOICE_SOLVER, lang,
                                       include_dialects)


def get_multiple_choice_solver_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.MULTIPLE_CHOICE_SOLVER)


def find_reading_comprehension_solver_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.READING_COMPREHENSION_SOLVER)


def load_reading_comprehension_solver_plugin(module_name: str) -> \
        type(EvidenceSolver):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.READING_COMPREHENSION_SOLVER)


def get_reading_comprehension_solver_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(
        PluginTypes.READING_COMPREHENSION_SOLVER)


def get_reading_comprehension_solver_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: dict configurations by language (if provided)
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    # solver plugins return {lang: [list of config dicts]}
    return load_plugin_configs(module_name,
                               PluginConfigTypes.READING_COMPREHENSION_SOLVER,
                               True)


def get_reading_comprehension_solver_lang_configs(lang: str,
                                     include_dialects: bool = False) -> dict:
    """
    Get a dict of plugin names to list valid configurations for the requested
    lang.
    @param lang: Language to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: `valid_configs`]}
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.READING_COMPREHENSION_SOLVER,
                                       lang, include_dialects)


def get_reading_comprehension_solver_supported_langs() -> dict:
    """
    Return a dict of plugin names to list supported languages
    @return: dict plugin names to list supported languages
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(
        PluginTypes.READING_COMPREHENSION_SOLVER)




from ovos_plugin_manager.utils import normalize_lang, \
    PluginTypes, PluginConfigTypes
from ovos_config import Configuration
from ovos_plugin_manager.utils.config import get_valid_plugin_configs, \
    sort_plugin_configs, get_plugin_config
from ovos_utils.log import LOG, log_deprecation
from ovos_plugin_manager.templates.stt import STT, StreamingSTT, StreamThread


def find_stt_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.STT)


def load_stt_plugin(module_name: str) -> type(STT):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.STT)


def get_stt_configs() -> dict:
    """
    Get a dict of plugin names to valid STT configuration
    @return: dict plugin name to dict of str lang to list of dict valid configs
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.STT)


def get_stt_module_configs(module_name: str) -> dict:
    """
    Get a dict of lang to list of valid config dicts for a specific plugin
    @param module_name: name of plugin to get configurations for
    @return: {lang: [list of config dicts]}
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    configs = load_plugin_configs(module_name, PluginConfigTypes.STT, True)
    # let's sort by priority key
    for k, v in configs.items():
        configs[k] = sorted(v, key=lambda c: c.get("priority", 60))
    return configs


def get_stt_lang_configs(lang: str, include_dialects: bool = False) -> dict:
    """
    Get a dict of plugins names to sorted list of valid configurations
    @param lang: language to get configurations for (i.e. en, en-US)
    @param include_dialects: If true, include configs for other locales
        (i.e. include en-GB configs for lang=en-US)
    @return: dict plugin name to list of valid configs sorted by priority
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    matched_configs = get_plugin_language_configs(PluginTypes.STT, lang,
                                                  include_dialects)
    return sort_plugin_configs(matched_configs)


def get_stt_supported_langs() -> dict:
    """
    Get a dict of languages to valid configuration options
    @return: dict lang to list of plugins that support that lang
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.STT)


def get_stt_config(config: dict = None, module: str = None) -> dict:
    """
    Get relevant configuration for factory methods
    @param config: global Configuration OR plugin class-specific configuration
    @param module: STT module to get configuration for
    @return: plugin class-specific configuration
    """
    from ovos_plugin_manager.utils.config import get_plugin_config
    stt_config = get_plugin_config(config, "stt", module)
    assert stt_config.get('lang') is not None, "expected lang but got None"
    return stt_config


class OVOSSTTFactory:
    """ replicates the base mycroft class, but uses only OPM enabled plugins"""

    @staticmethod
    def get_class(config=None):
        """Factory method to get a STT engine class based on configuration.

        The configuration file ``mycroft.conf`` contains a ``stt`` section with
        the name of a STT module to be read by this method.

        "stt": {
            "module": <engine_name>
        }
        """
        config = get_stt_config(config)
        stt_module = config["module"]
        return load_stt_plugin(stt_module)

    @staticmethod
    def create(config=None):
        """Factory method to create a STT engine based on configuration.

        The configuration file ``mycroft.conf`` contains a ``stt`` section with
        the name of a STT module to be read by this method.

        "stt": {
            "module": <engine_name>
        }
        """
        stt_config = get_stt_config(config)
        try:
            clazz = OVOSSTTFactory.get_class(stt_config)
            return clazz(stt_config)
        except Exception:
            LOG.exception('The selected STT plugin could not be loaded!')
            raise



import abc

from typing import Tuple, List, Iterable


class TriplesExtractor:

    def __init__(self, config=None):
        self.config = config or {}
        self.first_person_token = self.config.get("first_person_token", "USER")

    @abc.abstractmethod
    def extract_triples(self, documents: List[str]) -> Iterable[Tuple[str, str, str]]:
        """Extract semantic triples from a list of documents."""



from typing import Optional, Union
from ovos_config.config import Configuration
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.log import LOG
from ovos_plugin_manager.utils import load_plugin, find_plugins, PluginTypes, PluginConfigTypes
from langcodes import tag_distance


def get_plugin_config(config: Optional[dict] = None, section: str = None,
                      module: Optional[str] = None) -> dict:
    """
    Get a configuration dict for the specified plugin. Configuration is applied
    such that:
    - module-specific configurations take priority
    - section-specific configuration is appended (new keys only)
    - global `lang` configuration is appended (if not already set)
    @param config: Base configuration to parse, defaults to `Configuration()`
    @param section: Config section for the plugin (i.e. TTS, STT, language)
    @param module: Module/plugin to get config for, default reads from config
    @return: Configuration for the requested module, including `lang` and `module` keys
    """
    config = config or Configuration()
    lang = standardize_lang_tag(config.get('lang') or Configuration().get('lang', "en"))
    config = (config.get('intentBox', {}).get(section) or config.get(section)
              or config) if section else config
    module = module or config.get('module')
    if module:
        module_config = dict(config.get(module) or dict())
        module_config.setdefault('module', module)
        for key, val in config.items():
            # Configured module name is not part of that module's config
            if key in ("module", "translation_module", "detection_module"):
                continue
            elif isinstance(val, dict):
                continue
            # Use section-scoped config as defaults (i.e. TTS.lang)
            module_config.setdefault(key, val)
        config = module_config
    if section not in ["hotwords", "VAD", "listener", "gui"]:
        config.setdefault('lang', lang)
    LOG.debug(f"Loaded configuration: {config}")
    return config


def get_valid_plugin_configs(configs: dict, lang: str,
                             include_dialects: bool) -> list:
    """
    Get a sorted dict of configurations for a particular plugin
    @param configs: dict of normalized language to sorted list of valid dict
                    configurations for a particular plugin
    @param lang: normalized language to return valid configurations for
    @param include_dialects: if True, include configs for alternate dialects
    @return: list of valid configurations matching the requested lang
    """
    valid_configs = list()
    if include_dialects:
        # Check other dialects of the requested language
        base_lang = standardize_lang_tag(lang, macro=True)
        for language, confs in configs.items():
            if tag_distance(base_lang, language) < 10:
                for config in confs:
                    try:
                        if language != lang:
                            # Dialect match, boost priority
                            config["priority"] = config.get("priority",
                                                            60) + 15
                        valid_configs.append(config)
                    except Exception as e:
                        LOG.error(f'config={config}')
                        LOG.exception(e)
    elif lang in configs:
        # Exact language/dialog match
        valid_configs.append(configs[lang])
    elif f"{lang}-{lang}" in configs:
        # match (some) default locales
        valid_configs.append(configs[f"{lang}-{lang}"])
    LOG.debug(f'Found {len(valid_configs)} valid configurations for {lang}')
    return valid_configs


def sort_plugin_configs(configs: dict) -> dict:
    """
    Sort a dict of plugin name to valid configurations by priority
    @param configs: dict config name to valid configurations
    @return: dict of sorted lists with highest priority at the end of the list
    """
    bad_plugs = []
    for plug_name, plug_configs in configs.items():
        LOG.debug(plug_configs)
        try:
            configs[plug_name] = sorted(plug_configs,
                                        key=lambda c: c.get("priority", 60))
        except:
            LOG.exception(f"Invalid plugin data: {plug_name}")
            bad_plugs.append(plug_name)

    for plug_name in [p for p in bad_plugs if p in configs]:
        configs.pop(plug_name)

    LOG.debug(configs)
    return {k: v for k, v in configs.items() if v}


def load_plugin_configs(plug_name: str,
                        plug_type: Optional[PluginConfigTypes] = None,
                        normalize_language_keys: bool = False) -> \
        Union[dict, list]:
    """
    Load a specific plugin's valid configurations.

    Arguments:
        plug_type: (str) plugin type name. Ex. "mycroft.plugin.tts".
        plug_name: (str) specific plugin name
        normalize_language_keys: (bool) If true, normalize dict keys as langs
    Returns:
        Loaded configuration dict, list of dicts, or None
        if no matching object was found.
    """
    config = load_plugin(plug_name + ".config", plug_type)
    if normalize_language_keys:
        return {standardize_lang_tag(lang): v for lang, v in config.items()}
    return config


def load_configs_for_plugin_type(plug_type: PluginTypes) -> dict:
    """
    Load all valid configuration options for the specified plug_type
    @param plug_type: Plugin type to get configs for
    @return: dict plugin name to list or dict configurations
    """
    return {plug: load_plugin_configs(
            plug, PluginConfigTypes(f"{plug_type.value}.config"))
            for plug in find_plugins(plug_type)} or dict()


def get_plugin_supported_languages(plug_type: PluginTypes) -> dict:
    """
    Return a dict of plugin names to list supported languages
    @param plug_type: plugin type to get plugins/configuration for
    @return: dict plugin names to list supported languages
    """
    lang_configs = dict()
    for plug in find_plugins(plug_type):
        configs = load_plugin_configs(plug, PluginConfigTypes(f"{plug_type.value}.config")) or {}
        for lang, config in configs.items():
            lang = standardize_lang_tag(lang)
            lang_configs.setdefault(lang, list())
            lang_configs[lang].append(plug)
    return lang_configs


def get_plugin_language_configs(plug_type: PluginTypes, lang: str,
                                include_dialects: bool = False) -> dict:
    """
    Return a dict of plugin names to list of valid (dict) configurations
    @param plug_type: plugin type to get configurations for
    @param lang: BCP-47 language code to get configurations for
    @param include_dialects: consider configurations in different locales
    @return: dict {`plugin_name`: [`valid_configs`]}
    """
    lang = standardize_lang_tag(lang)
    plugin_configs = dict()
    valid_configs = dict()
    for plug in find_plugins(plug_type):
        plugin_configs[plug] = list()
        valid_configs = \
            load_plugin_configs(plug,
                                PluginConfigTypes(f"{plug_type.value}.config"))
        valid_configs = {standardize_lang_tag(lang): conf
                         for lang, conf in valid_configs.items()}
        if include_dialects:
            lang = standardize_lang_tag(lang, macro=True)
            for language in valid_configs:
                if language.startswith(lang):
                    plugin_configs[plug] += valid_configs[language]
        elif lang in valid_configs:
            plugin_configs[plug] += valid_configs[lang]
        elif f"{lang}-{lang}" in valid_configs:
            plugin_configs += valid_configs[f"{lang}-{lang}"]
    return {plug: configs for plug, configs in
            valid_configs.items() if configs} or dict()
